from zope.component.factory import Factory
from zope import interface
from . import interfaces as qry_ifaces

import logging
logger = logging.getLogger(__name__)

@interface.implementer(qry_ifaces.IORMRelatedModelsAdder)
class ORMRelatedModelsAdder(object):
    def __init__(self, models=None):
        self._models = {}
        if models:
            self.add_sequence(models)
    def add_sequence(self, sequence):
        for i, model in enumerate(sequence):
            if model not in self._models:
                self._models[model] = set()
            #add left-side
            if i:
                self._models[model].add(sequence[i-1])
            #add right-side
            if i < len(sequence)-1:
                self._models[model].add(sequence[i+1])
            
    def flattened(self, seed):
        q = [seed]
        flattened = []
        while q:
            model = q.pop(0)
            if model not in flattened:
                flattened.append(model)
            for r in [r for r in self._models[model] if r not in flattened]:
                q.append(r)
        return tuple(flattened)
ORMRelatedModelsAdderFactory = Factory(ORMRelatedModelsAdder)


@interface.implementer(qry_ifaces.ISAInstrumentedAttribute)
class SAInstrumentedAttributeFromDottedString(object):
    def __new__(self, Base, dotted_name):
        class_name, property_name = dotted_name.split('.')
        base_classes = Base._decl_class_registry
        if class_name not in base_classes:
            raise TypeError('unable to find {} type in SQLAlchemy base classes {}'.format(class_name, base_classes.keys()))
        class_ = base_classes[class_name]
        property_ = getattr(class_, property_name)
        interface.alsoProvides(property_, qry_ifaces.ISAInstrumentedAttribute)
        return getattr(class_, property_name)
SAInstrumentedAttributeFromDottedStringFactory = Factory(SAInstrumentedAttributeFromDottedString)
    

@interface.implementer(qry_ifaces.ISAQuery)
class SAOuterJoinQuery(object):
    def __new__(cls, session, models, select=None):
        """Return ISAQuery provider based on related models
        
        args:
            session: sparc.sa.orm.ISASession provider
            models: sequence of ISAModel providers ordered left->right in terms
                    of required SQL query dependency structure.
            select: sequence of models to return data for.  default is all 
                    entries in models
        """
        models = [m for m in models]
        if not select:
            select = models
        
        q = session.query(*select) # base query
        for m in models[1:]:
            q = q.outerjoin(m)
        logger.debug("ISAQuery provider inited with sequenced joins {}".format(models))
        return q
SAOuterJoinQueryFactory = Factory(SAOuterJoinQuery)




from zope import interface
from zope.component.factory import Factory

import sqlalchemy
from . import ISAEngine

@interface.implementer(ISAEngine)
class SAEngineFromConfig(object):
    def __new__(cls, SQLAlchemyEngine):
        """Return a ISAEngine provider
        
        Kwargs:
            see SQLAlchemyEngine def in configure.yaml
        """
        return sqlalchemy.create_engine(SQLAlchemyEngine.get('dsn', ''), **SQLAlchemyEngine.get('kwargs', {}))
SAEngineFromConfigFactory = Factory(SAEngineFromConfig)


from zope import interface

class ISADeclarativeMeta(interface.Interface):
    """Marker for sqlalchemy.ext.declarative.api.DeclarativeMeta"""
import sqlalchemy.ext.declarative.api
interface.classImplements(sqlalchemy.ext.declarative.api.DeclarativeMeta, ISADeclarativeMeta)


from zope.component.factory import Factory
from zope import component
from zope import interface
from zope.schema.fieldproperty import FieldProperty
from sqlalchemy import and_, or_
from . import interfaces as qry_ifaces
from . import query

@interface.implementer(qry_ifaces.ISAModelFilterExpression)
class SAModelFilterExpression(object):
    def __init__(self, *args, **kwargs):
        self.attribute = args[0] if len(args)>0 else kwargs['attribute']
        self.condition = args[1] if len(args)>1 else kwargs['condition']
        if len(args)>2 or 'value' in kwargs:
            self.value = args[2] if len(args)>2 else kwargs['value']
    condition = FieldProperty(qry_ifaces.ISAModelFilterExpression['condition'])
    value = FieldProperty(qry_ifaces.ISAModelFilterExpression['value'])
SAModelFilterExpressionFactory = Factory(SAModelFilterExpression)

@interface.implementer(qry_ifaces.ISAExpression)
@component.adapter(qry_ifaces.ISAModelFilterExpression)
class SAExpressionFromSAModelFilterExpression(object):
    def __new__(self, context):
        c = context.condition
        if c == '==' or c.lower() == 'equals':
            _return = context.attribute.__eq__(context.value)
        elif c == '!=' or c.lower() == 'not equals':
            _return = context.attribute.__ne__(context.value)
        elif c.lower() == 'like':
            _return = context.attribute.like(context.value)
        elif c.lower() == 'ilike':
            _return = context.attribute.ilike(context.value)
        elif c.lower() == 'in':
            _return = context.attribute.in_(context.value)
        elif c.lower() == 'not in':
            _return = ~context.attribute.in_(context.value)
        elif c.lower() == 'is null':
            _return = context.attribute.is_(None)
        elif c.lower() == 'is not null':
            _return = context.attribute.isnot(None)
        else:
            raise ValueError("received unexpected condition {}".format(c))
        interface.alsoProvides(_return, qry_ifaces.ISAExpression)
        return _return

@interface.implementer(qry_ifaces.ISAModelFilterExpressionGroup)
class SAModelFilterExpressionGroup(object):
    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            self.conjunction = args[0]
        if 'conjunction' in kwargs:
            self.conjunction = kwargs['conjunction']
        if len(args) > 1:
            self.expressions = args[1]
        if 'expressions' in kwargs:
            self.expressions = kwargs['expressions']
    conjunction = FieldProperty(qry_ifaces.ISAModelFilterExpressionGroup['conjunction'])
    expressions = FieldProperty(qry_ifaces.ISAModelFilterExpressionGroup['expressions'])
SAModelFilterExpressionGroupFactory = Factory(SAModelFilterExpressionGroup)

def resolve_expression_group(eg):
    """Recursive expression group resolver
    
    Args:
        eg: ISAModelFilterExpressionGroup provider
    
    Returns:
        SQLAlchemy conjunction for given expression group
    """
    conj = and_ if eg.conjunction.upper() == 'AND' else or_
    expressions = []
    for cond in eg.expressions:
        if qry_ifaces.ISAModelFilterExpressionGroup.providedBy(cond):
            expressions.append(resolve_expression_group(cond))
        else: #ISAModelFilterExpression
            expressions.append(qry_ifaces.ISAExpression(cond))
    return conj(*expressions)

@interface.implementer(qry_ifaces.ISAConjunction)
@component.adapter(qry_ifaces.ISAModelFilterExpressionGroup)
class SAConjunctionFromSAModelFilterExpressionGroup(object):
    
    def __new__(cls, context):
        conj = resolve_expression_group(context)
        interface.alsoProvides(conj, qry_ifaces.ISAConjunction)
        return conj



def is_expression_container(container):
    return 'attribute' in container and 'condition' in container

def is_expression_group_container(container):
    return 'conjunction' in container and 'expressions' in container

def convert_expression_group_container(Base, eg_container):
    """Recursive expression group container converter
    
    Args:
        eg_container: container with keys related to ISAModelFilterExpressionGroup containment
    
    Returns:
        ISAModelFilterExpressionGroup provider
    """
    conj = eg_container['conjunction'].upper()
    expressions = []
    for cond in eg_container['expressions']:
        if is_expression_group_container(cond):
            expressions.append(convert_expression_group_container(cond))
        else: #ISAModelFilterExpression container
            cond['attribute'] = query.SAInstrumentedAttributeFromDottedStringFactory(Base, cond['attribute'])
            expressions.append(SAModelFilterExpressionFactory(**cond))
    return SAModelFilterExpressionGroupFactory(
                                conjunction=conj, expressions=set(expressions))

@interface.implementer(qry_ifaces.ISAModelFilterExpressionGroup)
class SAModelFilterExpressionGroupFromContainer(object):
    def __new__(self, Base, container):
        """Return ISAModelFilterExpressionGroup based on formated container
        
        Container format should follow this paradigm:
            {
             'conjunction': 'and',
             'expressions':
                [
                 {'attribute':'model.field1', 'condition':'==', 'value':'value1'},
                 {'attribute':'model.field2', 'condition':'!=', 'value':'value2'},
                 {
                  'conjunction': 'or',
                  'expressions':
                      [
                       {'attribute':'model.field3', 'condition':'==', 'value':'value3'},
                       {'attribute':'model.field4', 'condition':'is null'}
                      ]
                 }
                ]
            }
        """
        return convert_expression_group_container(Base, container)
SAModelFilterExpressionGroupFromContainerFactory = Factory(SAModelFilterExpressionGroupFromContainer)



from zope import interface
from zope import component
from sqlalchemy import orm

from contextlib import contextmanager
from . import ISAScopedTransaction
from ..engine import ISAEngine
from ..ext import ISADeclarativeMeta

@interface.implementer(ISAScopedTransaction)
@component.adapter(ISAEngine, ISADeclarativeMeta)
class ThreadLocalSASessionTransaction(object):
    """Thread safe session transaction context manager implementation
    
    Args:
        engine: ISAEngine provider
        base: ISADeclarativeMeta provider
    """
    def __init__(self, engine, base):
        self._engine = engine
        self._Base = base
        
        self._session_factory = orm.sessionmaker(bind=self.engine)
        self._Session = orm.scoped_session(self._session_factory)
    
    @property
    def Base(self):
        return self._Base
    
    @property
    def engine(self):
        return self._engine
    
    @property
    def session(self):
        return self._Session() #thread-local session
    
    @contextmanager
    def __call__(self):
        try:
            yield self._Session
            self.session.commit()
        except:
            self.session.rollback()
            raise
        finally:
            self._Session.remove() #http://docs.sqlalchemy.org/en/latest/orm/contextual.html



import os
import zope.testrunner
from sparc.testing.fixture import test_suite_mixin
from sparc.sa.testing import SPARC_SA_INTEGRATION_LAYER


class test_suite(test_suite_mixin):
    package = 'sparc.sa.orm'
    module = 'session'
    layer = SPARC_SA_INTEGRATION_LAYER


if __name__ == '__main__':
    zope.testrunner.run([
                         '--path', os.path.dirname(__file__),
                         '--tests-pattern', os.path.splitext(
                                                os.path.basename(__file__))[0]
                         ])


from sparc.sa.engine import engine
from sparc.sa.orm.session import ThreadLocalSASessionTransaction
from sqlalchemy.pool import StaticPool

#Our declarative Base
from sqlalchemy.ext.declarative import declarative_base
Base = declarative_base()

def get_transaction():
    """Return ISAScopedTransaction provider for sqlite memory db with testing 
       schema initialized
    """
    #Memory sqlite engine that isn't thread-safe...but will work with threads
    #See http://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#pysqlite-threading-pooling
    sa_engine = engine.SAEngineFromConfigFactory({'dsn': 'sqlite:///:memory:',
                        'kwargs': {'connect_args':{'check_same_thread':False},
                        'poolclass':StaticPool}})
    transaction = ThreadLocalSASessionTransaction(sa_engine, Base)
    transaction.Base.metadata.create_all(bind=sa_engine)
    return transaction

def populate_tables(transaction):
    create_model(transaction, Test1, **{'name': 'test1_1'})
    create_model(transaction, Test1, **{'name': 'test1_2'})
    create_model(transaction, Test1, **{'name': 'test1_3'})
    
    create_model(transaction, Test2, **{'name': 'test2_1'})
    create_model(transaction, Test2, **{'name': 'test2_2'})
    create_model(transaction, Test2, **{'name': 'test2_3'})
    
    create_model(transaction, Test3, **{'test1_id': 'test1_1',
                                      'test2_id': 'test2_1'})
    create_model(transaction, Test3, **{'test1_id': 'test1_1',
                                      'test2_id': 'test2_2'})
    
    create_model(transaction, Test4, **{'name': 'test4_1',
                                      'test2_id': 'test2_1'})
    
    create_model(transaction, Test4, **{'name': 'test4_2',
                                      'test2_id': 'test2_3'})

def create_model(transaction, class_, **attributes):
    session = transaction.session
    model = class_()
    for k in attributes:
        setattr(model, k, attributes[k])
    session.add(model)
    session.flush()
    return model

#A semi-complex db schema with relationships
import sqlalchemy
from zope import interface
class ITest1(interface.Interface):
    """Marker"""
@interface.implementer(ITest1)
class Test1(Base):
    __tablename__ = 'test1'
    id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False, primary_key=True, autoincrement=True)
    name = sqlalchemy.Column(sqlalchemy.String, nullable=True, unique=False)

class ITest2(interface.Interface):
    """Marker"""
@interface.implementer(ITest2)
class Test2(Base):
    __tablename__ = 'test2'
    id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False, primary_key=True, autoincrement=True)
    name = sqlalchemy.Column(sqlalchemy.String, nullable=True, unique=False)

class ITest3(interface.Interface):
    """Marker"""
@interface.implementer(ITest3)
class Test3(Base):
    __tablename__ = 'test3'
    __table_args__ = \
                    (sqlalchemy.PrimaryKeyConstraint(\
                            'test2_id','test1_id', 
                            name='test3_pk'),
                     )
    test1_id = \
                sqlalchemy.Column(sqlalchemy.String, 
                    sqlalchemy.ForeignKey(\
                                    Test1.__tablename__ + '.id'),
                    nullable=False)
                
    test2_id = sqlalchemy.Column(sqlalchemy.Integer, 
                    sqlalchemy.ForeignKey(Test2.__tablename__ + '.id'),
                    nullable=False)


class ITest4(interface.Interface):
    """Marker"""
@interface.implementer(ITest4)
class Test4(Base):
    __tablename__ = 'test4'
    __table_args__ = \
                    (sqlalchemy.UniqueConstraint(\
                            'name','test2_id', name='test4_name_un'),
                     )
    id = sqlalchemy.Column(sqlalchemy.Integer, nullable=False, primary_key=True, autoincrement=True)
    name = sqlalchemy.Column(sqlalchemy.String, nullable=False)
    test2_id = sqlalchemy.Column(sqlalchemy.Integer, 
                    sqlalchemy.ForeignKey(Test2.__tablename__ + '.id'),
                    nullable=False)


from zope import interface

class ISAInstrumentedAttribute(interface.Interface):
    """SQLAlchemy ORM model attribute"""
from sqlalchemy.orm import attributes
interface.classImplements(attributes.InstrumentedAttribute, ISAInstrumentedAttribute)



from zope import interface

class ISASession(interface.Interface):
    """Marker for sqlalchemy.orm.session.Session"""
import sqlalchemy.orm.session
interface.classImplements(sqlalchemy.orm.session.Session, ISASession)

class ISAScopedSession(interface.Interface):
    """Marker for sqlalchemy.orm.scoping.scoped_session"""
import sqlalchemy.orm.scoping
interface.classImplements(sqlalchemy.orm.scoping.scoped_session, ISAScopedSession)

class ISAScopedTransaction(interface.Interface):
    """SQLAlchemy context manager for sessions"""
    Base = interface.Attribute("ISADeclarativeMeta provider")
    engine = interface.Attribute("ISAEngine provider")
    session = interface.Attribute("ISASession provider")
    
    def __call__():
        """Return a Python context manager whose __enter__() returns a 
        ISAScopedSession provider and whose __exit__() will commit/abort the
        current session transaction"""
    


from .interfaces import ISAConjunction
from .interfaces import ISAExpression
from .interfaces import ISAInstrumentedAttribute
from .interfaces import ISAModelFilterExpression
from .interfaces import ISAModelFilterExpressionGroup
from .interfaces import IORMRelatedModelsAdder
from .interfaces import IORMRelatedModels
from .interfaces import ISAQuery
from .interfaces import ISAResultRowModel


from zope import interface
from zope import component
from zope.interface.common.mapping import IIterableMapping
from .. import ISASession
from . import ISAInstrumentedAttribute

@interface.implementer(IIterableMapping)
@component.adapter(ISAInstrumentedAttribute, ISASession)
class ModelMappingForUniqueInstrumentedAttribute(object):
    """Read Mapping of sparc.sa.ext.ISADeclarativeMeta providers with 
       sparc.sa.attributes.ISAInstrumentedAttribute scalers as keys
    
    This effectively creates a map of a column value vs the matched row.
    """
    
    def __init__(self, attribute, session):
        self.attribute = attribute
        self.session = session
        
        self.attribute_model_type = attribute.class_
        self.attribute_name = attribute.property.key
        self.query = self.session.query(self.attribute_model_type)
    
    #IItemMapping
    def __getitem__(self, key):
        """Get a value for a key

        A KeyError is raised if there is no value for the key.
        """
        model = self.query.filter(self.attribute == key).one_or_none()
        if not model:
            raise KeyError(key)
        return model
    
    #IReadMapping
    def get(self, key, default=None):
        """Get a value for a key

        The default is returned if there is no value for the key.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        """Tell if a key exists in the mapping."""
        return True if self.query.filter(self.attribute == key).one_or_none() else False
    
    
    #IEnumerableMapping  
    def keys(self):
        """Return the keys of the mapping object.
        """
        return [key for key in self]

    def __iter__(self):
        """Return an iterator for the keys of the mapping object.
        """
        for model in self.query:
            yield getattr(model, self.attribute_name)

    def values(self):
        """Return the values of the mapping object.
        """
        return list(self.itervalues())

    def items(self):
        """Return the items of the mapping object.
        """
        return list(self.iteritems())

    def __len__(self):
        """Return the number of items.
        """
        return self.query.count()

    #IIterableMapping
    def iterkeys(self):
        "iterate over keys; equivalent to __iter__"
        return iter(self)

    def itervalues(self):
        "iterate over values"
        for model in self.query:
            yield model

    def iteritems(self):
        "iterate over items"
        for model in self.query:
            yield (getattr(model, self.attribute_name), model)



from zope import interface
from . import ISAResultRowModel

@interface.implementer(ISAResultRowModel)
class SAResultRowModel(object):
    def first(self, iface, row):
        for col in row:
            if iface.providedBy(col):
                return col
        raise ValueError("unable to find provider for {} in collection {}".format(iface, row))


from zope import interface
from zope import schema

class IORMRelatedModels(interface.Interface):
    def flattened(seed):
        """Return non-circular sequence of IORMModel providers from seed starting point"""

class IORMRelatedModelsAdder(IORMRelatedModels):
    def add_sequence(sequence):
        """Add sequence of related IORMModel providers into graph"""

class ISAQuery(interface.Interface):
    """SQLAlchemy query object initialized with desired return items"""
from sqlalchemy.orm.query import Query
interface.classImplements(Query, ISAQuery)

class ISAExpression(interface.Interface):
    """SQLAlchemy query filter expression that can be used within a conjunction"""

class ISAInstrumentedAttribute(interface.Interface):
    """SQLAlchemy ORM model attribute"""

class ISAConjunction(interface.Interface):
    """Result from SQLAlchemy and_() or or_(), object represents conjunction of expressions"""

class ISAResultRowModel(interface.Interface):
    """SQLAlchemy model query result row parser"""
    def first(iface, row):
        """Return first result row model entry providing given interface
        
        Args:
            iface: zope.interface based specification
            row: sqlachemy query result row
        """

class ISAModelFilterExpression(interface.Interface):
    attribute = interface.Attribute(u"ISAModel attribute to filter by")
    condition = schema.Choice(
                title=u"Filter conditional matching logic",
                values=['==','equals','!=','not equals','like','ilike',
                        'in','not in','is null', 'is not null'],
                required=True
            )
    value = schema.Field(
                title=u"value to be used against filter condition",
                required=False
            )

class ISAModelFilterExpressionGroup(interface.Interface):
    conjunction = schema.Choice(
                title=u"Append logic for condition statements within the group",
                values=['AND','OR']
            )
    expressions = schema.Set(
                title=u"Group of expressions and sub-expression groups"
            )
